section1_dup1=val1
section1_dup2=val2
section1_nospace=val
section1_multiline='with
leading
space'
section1_nmultiline='not supported with\'
section1_comment_after1=val
section1_comment_after2='val;not a comment'
section1_comment_after3='val #not a comment'
section1_escaped_not_processed='test \nescape'
section1_colon=val
section1_double_quotes='"not removed"'
section1_single_quotes=''"'"'not removed'"'"''
section1_spaces_stripped=val
section1_internal_not_stripped='v  al'
section1_notempty1=';comment=val'
section1_empty=''
section1_python_interpolate='%(dup1)s/blah'
section1_interpolate2='${dup1}/blah'
section1_Caps='not significant'
section1_combine=sections
